(function() {
 var Constants = {
  ALLOWED_CALC_WIDTH_DIFF: 0.5,
  ALLOWED_RENDER_WIDTH_DIFF: 0.01,
  MAX_ITERATION_COUNT: 1000
 };
 var GridViewColumnAdjustor = ASPx.CreateClass(null, {
  constructor: function(grid) {
   this.grid = grid;
   this.columnInfos = [];
   this.hasRestrictions = false;
   this.hasValidColumnWidths = true;
  },
  Initialize: function() {
   this.columnInfos = this.GetColumnInfos();
   this.hasRestrictions = this.CheckHasRestrictions(this.columnInfos);
   this.hasValidColumnWidths = this.IsValidColumnWidths(this.columnInfos);
  },
  GetColumnInfos: function() {
   var columnInfos = [];
   var armRow = this.GetMainTableArmRow();
   var columns = this.grid.columns;
   for(var i = 0, column; column = columns[i]; i++) {
    if(!column.visible)
     continue;
    var cellIndex = this.grid.GetDataCellIndex(i);
    if(cellIndex === -1)
     continue;
    var armCell = armRow.cells[cellIndex];
    if(!armCell || armCell.colSpan > 1)
     continue;
    var styleWidth = armCell.style.width;
    var hasPercentageWidth = false;
    var hasFixedWidth = false;
    var width = -1;
    if(styleWidth) {
     hasPercentageWidth = ASPx.IsPercentageSize(styleWidth);
     hasFixedWidth = !hasPercentageWidth;
     width = hasPercentageWidth ? ASPx.PercentageToFloat(styleWidth) : ASPx.PxToInt(styleWidth);
    }
    var info = this.CreateColumnInfoObject(i, column, hasPercentageWidth, hasFixedWidth, width);
    columnInfos.push(info);
   }
   return columnInfos;
  },
  CheckHasRestrictions: function(columnInfos) {
   for(var info, i = 0; info = columnInfos[i]; i++) {
    if(info.minWidth > 0 || info.maxWidth > 0)
     return true;
   }
   return false;
  },
  HasRestrictions: function() {
   return this.hasRestrictions;
  },
  IsValidColumnWidths: function(columnInfos) {
   var hasPercWidths = false,
    hasAutoWidths = false;
   for(var info, i = 0; info = columnInfos[i]; i++) {
    if(info.hasPercentageWidth) {
     if(hasAutoWidths)
      return false;
     hasPercWidths = true;
    }
    else if(!info.hasFixedWidth) {
     if(hasPercWidths)
      return false;
     hasAutoWidths = true;
    }
   }
   return true;
  },
  HasValidColumnWidths: function() {
   return this.hasValidColumnWidths;
  },
  ShouldAdjustColumns: function() {
   return this.HasRestrictions() && this.HasValidColumnWidths();
  },
  IsColumnAdjustmentAllowed: function() {
   return !this.grid.AllowResizing() && !this.grid.IsResponsiveAdaptivityMode();
  },
  OnAfterCallback: function() {
   this.Initialize();
   this.AdjustColumns();
  },
  AdjustColumns: function() {
   if(!this.ShouldAdjustColumns() || !this.IsColumnAdjustmentAllowed())
    return;
   var savedStyle = this.CollapseTables();
   var gridWidth = this.GetGridOffsetWidth(); 
   this.AdjustColumnsCore(this.columnInfos, gridWidth);
   this.ExpandTables(savedStyle);
  },
  AdjustColumnsCore: function(columnInfos, gridWidth) {
   this.ResetColumnViolations(columnInfos);
   var resolvedGridInfo = this.CalculateCompleteResolvedColumnInfos(columnInfos, gridWidth);
   this.ApplyColumnInfosToRender(resolvedGridInfo.columnInfos, resolvedGridInfo.maxWidth);
  },
  ResetColumnViolations: function(columnInfos) {
   for(var info, i = 0; info = columnInfos[i]; i++)
    info.violation = null;
  },
  GetModifiedColumnInfos: function(newColumnInfos, oldColumnInfos) {
   var result = [];
   for(var newInfo, i = 0; newInfo = newColumnInfos[i]; i++) {
    oldInfo = oldColumnInfos[i];
    if(!this.AreEqualByWidth(newInfo, oldInfo))
     result.push(newInfo);
   }
   return result;
  },
  CalculateCompleteResolvedColumnInfos: function(sourceColumnInfos, gridWidth) {
   var violation = null;
   var resolvedGridInfo = null;
   var counter = 0;
   do {
    resolvedGridInfo = this.CalculateResolvedGridInfo(gridWidth, sourceColumnInfos);
    var resolvedGridWidth = ASPx.IsExists(resolvedGridInfo.resolvedWidth) ? resolvedGridInfo.resolvedWidth : gridWidth;
    var widthInfos = this.CalculateColumnsCurrentWidths(resolvedGridWidth, resolvedGridInfo.columnInfos);
    violation = this.GetWidthViolation(widthInfos);
    if(violation)
     this.ResolveViolation(violation);
    if(++counter > Constants.MAX_ITERATION_COUNT)
     throw "GridView column calculation error.";
   } while(violation);
   return resolvedGridInfo;
  },
  CollapseTables: function() {
   var styles = [];
   var cell = this.GetMainElementCell();
   var style = cell.style;
   var savedStyle = { display: style.display, width: style.width, height: style.height, overflow: style.overflow };
   ASPx.SetStyles(cell, { display: "block", width: "0px", height: "0px", overflow: "hidden", maxWidth: "" });
   styles.push(savedStyle);
   if(this.grid.widthValueSetInPercentage) {
    var mainElement = this.grid.GetMainElement();
    savedStyle = { width: mainElement.style.width };
    mainElement.style.width = this.grid.percentSizeConfig.width;
    styles.push(savedStyle);
   }
   return styles;
  },
  ExpandTables: function(styles) {
   var cell = this.GetMainElementCell();
   ASPx.SetStyles(cell, styles[0]);
   if(this.grid.widthValueSetInPercentage)
    this.grid.GetMainElement().style.width = styles[1].width;
  },
  RestoreTableRenderColumnWidths: function(table) {
   this.ApplyColumnInfosToTable(table, this.columnInfos);
  },
  ApplyColumnInfosToTable: function(table, columnInfos) {
   var armRow = table.rows[0];
   for(var info, i = 0; info = columnInfos[i]; i++) {
    if(!info.column) continue;
    var widthStr = this.GetColumnInfoWidthString(info);
    var columnIndex = info.column.index;
    var cellIndex = this.grid.GetDataCellIndex(columnIndex);
    var armCell = armRow.cells[cellIndex];
    armCell.style.width = widthStr;
   }
  },
  ApplyColumnInfosToRender: function(columnInfos, maxGridWidth) {
   var tables = this.grid.GetLayoutHelper().GetAllTables();
   for(var table, i = 0; table = tables[i]; i++)
    this.ApplyColumnInfosToTable(table, columnInfos);
   if(ASPx.IsExists(maxGridWidth))
    this.grid.GetMainElement().style.maxWidth = maxGridWidth + "px";
  },
  ResolveViolation: function(violation) {
   violation.columnInfo.violation = violation;
  },
  GetWidthViolation: function(widthInfos) {
   var minViolation = null,
    maxViolation = null,
    summaryWidthDiff = 0;
   for(var info, i = 0; info = widthInfos[i]; i++) {
    var columnInfo = info.columnInfo.sourceColumnInfo || info.columnInfo;
    if(columnInfo.minWidth > 0) {
     var widthDiff = columnInfo.minWidth - info.width;
     if(widthDiff > Constants.ALLOWED_CALC_WIDTH_DIFF) {
      summaryWidthDiff += widthDiff;
      if(!minViolation)
       minViolation = this.CreateViolationObject(true, columnInfo, widthDiff, columnInfo.minWidth);
      continue;
     }
    }
    if(columnInfo.maxWidth > 0) {
     var widthDiff = info.width - columnInfo.maxWidth;
     if(widthDiff > Constants.ALLOWED_CALC_WIDTH_DIFF) {
      summaryWidthDiff -= widthDiff;
      if(!maxViolation)
       maxViolation = this.CreateViolationObject(false, columnInfo, -widthDiff, columnInfo.maxWidth);
      continue;
     }
    }
   }
   return summaryWidthDiff >= 0 ? minViolation : maxViolation;
  },
  CreateViolationObject: function(minWidthViolation, columnInfo, widthDiff, violatedWidth) {
   return {
    minWidthViolation: minWidthViolation,
    columnInfo: columnInfo,
    widthDiff: widthDiff,
    violatedWidth: violatedWidth
   };
  },
  CalculateResolvedGridInfo: function(gridWidth, columnInfos) {
   var groupedInfos = this.GroupInfosByWidthType(columnInfos);
   var autoWidthInfos = groupedInfos.autoWidthInfos,
    fixedWidthInfos = groupedInfos.fixedWidthInfos,
    percentageWidthInfos = groupedInfos.percentageWidthInfos;
   if(autoWidthInfos.length > 0 && percentageWidthInfos.length == 0) {
    return this.CalculateResolvedAutoWidthColumnsGridInfo(gridWidth, columnInfos);
   }
   if(autoWidthInfos.length == 0 && fixedWidthInfos.length > 0 && percentageWidthInfos.length == 0) {
    return this.CalculateResolvedFixedWidthColumnsGridInfo(gridWidth, columnInfos);
   }
   if(autoWidthInfos.length == 0 && percentageWidthInfos.length > 0) {
    return this.CalculateResolvedPercentageWidthColumnsGridInfo(gridWidth, columnInfos);
   }
   return null;
  },
  CalculateResolvedPercentageWidthColumnsGridInfo: function(gridWidth, columnInfos) {
   var result = { columnInfos: [] };
   var fixedColumnCount = 0;
   var fixedWidthSummary = 0;
   var percWidthSummary = 0;
   for(var info, i = 0; info = columnInfos[i]; i++) {
    if(info.hasFixedWidth || info.violation) {
     fixedColumnCount++;
     fixedWidthSummary += info.hasFixedWidth ? info.width : info.violation.violatedWidth;
    }
    else
     percWidthSummary += info.width;
   }
   var percColumnSummaryWidth = gridWidth - fixedWidthSummary;
   var percColumnCount = columnInfos.length - fixedColumnCount;
   if(percColumnSummaryWidth > 0 && percColumnCount == 0)
    result.maxWidth = result.resolvedWidth = fixedWidthSummary;
   if(percWidthSummary == 0)
    percWidthSummary = 1;
   for(var info, i = 0; info = columnInfos[i]; i++) {
    var hasFixedWidth = true;
    var width = info.width;
    if(info.violation)
     width = info.violation.violatedWidth;
    else if(!info.hasFixedWidth) {
     hasFixedWidth = false;
     width = info.width / percWidthSummary;
    }
    var columnInfo = this.CreateColumnInfoObject(info.index, info.column, !hasFixedWidth, hasFixedWidth, width, info);
    result.columnInfos.push(columnInfo);
   }
   return result;
  },
  CalculateResolvedFixedWidthColumnsGridInfo: function(gridWidth, columnInfos) {
   var result = { columnInfos: [] };
   var violatedWidthSummary = 0;
   var violatedColumnCount = 0;
   var normalColumnMinSummaryWidth = 0;
   for(var info, i = 0; info = columnInfos[i]; i++) {
    if(info.violation) {
     violatedWidthSummary += info.violation.violatedWidth;
     violatedColumnCount++;
    }
    else
     normalColumnMinSummaryWidth += info.width;
   }
   var newMinGridWidth = violatedWidthSummary + normalColumnMinSummaryWidth;
   if(gridWidth <= newMinGridWidth) {
    for(var info, i = 0; info = columnInfos[i]; i++) {
     var columnInfo = this.CreateColumnInfoObject(info.index, info.column, false, true, info.width, info);
     result.columnInfos.push(columnInfo);
    }
    return result;
   }
   else if(violatedColumnCount == columnInfos.length)
    result.maxWidth = result.resolvedWidth = gridWidth = violatedWidthSummary;
   var normalColumnSummaryWidth = gridWidth - violatedWidthSummary;
   var normalColumnSummaryPart = normalColumnSummaryWidth / gridWidth;
   var violatedColumnSummaryPart = 1 - normalColumnSummaryPart;
   for(var info, i = 0; info = columnInfos[i]; i++) {
    var part = info.violation
     ? (info.violation.violatedWidth / violatedWidthSummary) * violatedColumnSummaryPart
     : (info.width / normalColumnMinSummaryWidth) * normalColumnSummaryPart;
    var width = part * gridWidth;
    var columnInfo = this.CreateColumnInfoObject(info.index, info.column, false, true, width, info);
    result.columnInfos.push(columnInfo);
   }
   return result;
  },
  CalculateResolvedAutoWidthColumnsGridInfo: function(gridWidth, columnInfos) {
   var result = { columnInfos: [] };
   var fixedColumnCount = 0;
   var fixedWidthSummary = 0;
   for(var info, i = 0; info = columnInfos[i]; i++) {
    if(info.hasFixedWidth || info.violation) {
     fixedColumnCount++;
     fixedWidthSummary += info.hasFixedWidth ? info.width : info.violation.violatedWidth;
    }
   }
   var autoColumnSummaryWidth = gridWidth - fixedWidthSummary;
   var autoColumnCount = columnInfos.length - fixedColumnCount;
   var autoWidth = 0;
   if(autoColumnSummaryWidth > 0) {
    if(autoColumnCount > 0)
     autoWidth = autoColumnSummaryWidth / autoColumnCount;
    else
     result.maxWidth = result.resolvedWidth = fixedWidthSummary;
   }
   for(var info, i = 0; info = columnInfos[i]; i++) {
    var hasFixedWidth = !!info.violation || info.hasFixedWidth;
    var width = autoWidth;
    if(info.hasFixedWidth)
     width = info.width;
    else if(info.violation)
     width = info.violation.violatedWidth;
    var columnInfo = this.CreateColumnInfoObject(info.index, info.column, false, hasFixedWidth, width, info);
    result.columnInfos.push(columnInfo);
   }
   return result;
  },
  CalculateColumnsCurrentWidths: function(gridWidth, columnInfos) {
   var groupedInfos = this.GroupInfosByWidthType(columnInfos);
   var autoWidthInfos = groupedInfos.autoWidthInfos,
    fixedWidthInfos = groupedInfos.fixedWidthInfos,
    percentageWidthInfos = groupedInfos.percentageWidthInfos;
   var result = [];
   if(percentageWidthInfos.length > 0 && autoWidthInfos.length > 0)
    return null;
   var fixedWidthSum = 0;
   for(var info, i = 0; info = fixedWidthInfos[i]; i++)
    fixedWidthSum += info.width;
   if(fixedWidthInfos.length == columnInfos.length) {
    var coefficient = 1;
    if(fixedWidthSum > 0)
     coefficient = gridWidth / fixedWidthSum;
    if(coefficient < 1)
     coefficient = 1;
    for(var info, i = 0; info = fixedWidthInfos[i]; i++) {
     var width = coefficient * info.width;
     var widthInfo = this.CreateWidthInfoObject(width, info);
     result.push(widthInfo);
    }
   }
   else {
    var flexWidthSum = gridWidth - fixedWidthSum;
    var autoWidthColumnWidth = 0;
    if(autoWidthInfos.length > 0 && flexWidthSum > 0)
     autoWidthColumnWidth = flexWidthSum / autoWidthInfos.length;
    var percWidthColumnSum = 0;
    for(var info, i = 0; info = percentageWidthInfos[i]; i++)
     percWidthColumnSum += info.width;
    if(percWidthColumnSum == 0)
     percWidthColumnSum = 1;
    for(var info, i = 0; info = columnInfos[i]; i++) {
     var width = info.width;
     if(info.hasPercentageWidth)
      width = flexWidthSum * info.width / percWidthColumnSum;
     else if(!info.hasFixedWidth)
      width = autoWidthColumnWidth;
     var widthInfo = this.CreateWidthInfoObject(width, info);
     result.push(widthInfo);
    }
   }
   return result;
  },
  GetColumnInfoWidthString: function(columnInfo) {
   if(columnInfo.hasFixedWidth)
    return columnInfo.width + "px";
   else if(columnInfo.hasPercentageWidth)
    return (columnInfo.width * 100) + "%";
   else
    return "";
  },
  AreEqualByWidth: function(columnInfo1, columnInfo2) {
   return columnInfo1.hasFixedWidth === columnInfo2.hasFixedWidth &&
    columnInfo1.hasPercentageWidth === columnInfo2.hasPercentageWidth &&
    (!columnInfo1.hasFixedWidth && !columnInfo1.hasPercentageWidth ||
     Math.abs(columnInfo1.width - columnInfo2.width) < Constants.ALLOWED_RENDER_WIDTH_DIFF);
  },
  CreateWidthInfoObject: function(width, columnInfo) {
   return {
    width: width,
    columnInfo: columnInfo
   };
  },
  CreateColumnInfoObject: function(index, column, hasPercentageWidth, hasFixedWidth, width, sourceColumnInfo) {
   return {
    index: index,
    column: column,
    hasPercentageWidth: hasPercentageWidth,
    hasFixedWidth: hasFixedWidth,
    width: width,
    minWidth: column ? column.minWidth : 0,
    maxWidth: column ? column.maxWidth : 0,
    sourceColumnInfo: sourceColumnInfo
   };
  },
  GroupInfosByWidthType: function(infos) {
   var autoWidthInfos = [],
    fixedWidthInfos = [],
    percentageWidthInfos = [],
    violatedWidthInfos = [];
   for(var i = 0, info; info = infos[i]; i++) {
    if(info.hasFixedWidth)
     fixedWidthInfos.push(info);
    else if(info.hasPercentageWidth)
     percentageWidthInfos.push(info);
    else
     autoWidthInfos.push(info);
    if(info.violation)
     violatedWidthInfos.push(info);
   }
   return {
    autoWidthInfos: autoWidthInfos,
    fixedWidthInfos: fixedWidthInfos,
    percentageWidthInfos: percentageWidthInfos,
    violatedWidthInfos: violatedWidthInfos
   };
  },
  GetGridOffsetWidth: function() {
   return this.grid.GetMainElement().offsetWidth;
  },
  GetMainTableArmRow: function () {
   return this.grid.GetMainTable().rows[0];
  },
  GetMainElementCell: function() {
   return this.grid.GetLayoutHelper().GetMainElementCell();
  }
 });
 ASPx.GridViewColumnAdjustor = GridViewColumnAdjustor;
})();
